#pragma once

#include "ublock.h"

enum dm_type {DMT_ELEMENT,DMT_ATTRIBUTE,DMT_ATTVAL,DMT_CHARS,DMT_INTENT,
	DMT_ADDKEY,DMT_ADDKEYVAL,DMT_NAMESPACE,DMT_CORPUSTEXT,DMT_MAX};

struct dm_declaration {
	UnicodeString key;
	int freq;
	dm_type ty;
	dm_declaration* next;
	dm_declaration():freq(1) {}
};

struct dm_element : public dm_declaration {
	UnicodeString e, uri, gloss; bool bBib;
};

struct dm_attribute : public dm_declaration {
	UnicodeString e, uri, a, uria, gloss, target;
	bool bG; bool bCB; bool bBib;
	indexpol ip; int ncb; _code* cb;
	int nv;
};

struct dm_addkey : public dm_declaration {
	UnicodeString k, gloss;
	bool bCB; bool isLemma, isPseudo; UnicodeString att;
	int ngi; UnicodeString* gi;
	UnicodeString defval; _addkey::defproc dproc;
	int ncb; _code* cb; int nv; int idx;
};

struct dm_intent : public dm_declaration {
	UnicodeString e, v;
};

struct dm_namespace : public dm_declaration {
	UnicodeString uri;
};

struct dm_corpustext : public dm_declaration {
	UnicodeString name;
	int* classes;
	UnicodeString* elts;
	UnicodeString* atts;
	int w;
	int* nr;
};

struct dm_value : public dm_declaration
{
	dm_attribute* a;
	UnicodeString v, gloss;
};

struct dm_addkeyvalue : public dm_declaration
{
	dm_addkey* a;
	UnicodeString v, gloss;
	int n;
};

class CDeclarationManager
{
public:
	CDeclarationManager(void);
	~CDeclarationManager(void);
	void DeclareElement(UnicodeString e,UnicodeString uri);
	void DeclareAddkey(UnicodeString,UnicodeString);
	void DeclareAttribute(UnicodeString e, UnicodeString,UnicodeString,UnicodeString,UnicodeString,bool);
	void DeclareCharacters(UnicodeString c);
	void DeclareInternalEntity(UnicodeString e, UnicodeString v);
	void DeclareAttributeValue(UnicodeString v);
	void DeclareNamespace(UnicodeString);
	void DeclareCorpusText(UnicodeString,int,int*);
	void DeclareClass(int,int,UnicodeString,UnicodeString);
	void SetHeader(class CSaraHeaderReader* h) {m_header=h;}
	void Add(dm_declaration*);
	void SetInBib(bool b) {m_bInBib=b;}
	UnicodeString CorpusTextName(UnicodeString);

private:
	bool m_bInBib;
	class CSaraHeaderReader* m_header;
	dm_declaration* m_list;
	dm_declaration** m_tail;
	static int elementcomp(const void*,const void*);
	static int valcomp(const void*,const void*);
	static int addkeyvalcomp(const void*,const void*);
	void DSCLabels(FILE*);
	void GrammarLabels(DOMElement*);
	void DSCAddKeys(FILE*);
	void GrammarAddKeys(DOMElement*);
	void DSCElements(FILE*);
	void GrammarNamespaces(DOMElement*);
	void GrammarElements(DOMElement*);
	void DSCBibElements(FILE*);
	void GrammarBibElements(DOMElement*);
	void DSCRanges(FILE*);
	void GrammarRanges(DOMElement*);
	void DSCAttribute(FILE*,dm_attribute*);
	void GrammarAttribute(DOMElement*,dm_attribute*);
	void DSCInternalEntities(FILE*);
	void GrammarInternalEntities(DOMElement*);
	void DSCValues(FILE*,dm_attribute*);
	void GrammarValues(DOMElement*,dm_attribute*);
	void DSCAddkeyValues(FILE*,dm_addkey*);
	void GrammarAddkeyValues(DOMElement*,dm_addkey*);
	void DSCTaxonomy(FILE*,dm_attribute*);
	void GrammarTaxonomy(DOMElement*,UnicodeString);
	void DSCTaxonomies(FILE*);
	void GrammarTaxonomies(DOMElement*);
	void DSCWBRules(FILE*);
	void DSCTexts(FILE*);
	void GrammarLemmaSchemes(DOMElement*);
	void DSCLemmaSchemes(FILE*);
	void DSCRegions(FILE*);
	void GrammarRegions(DOMElement*);
	bool IsClassTag(int,UnicodeString,UnicodeString);
	XERCES_CPP_NAMESPACE::DOMDocument* m_doc;
	void Ident(DOMElement*);
	int m_id;
	dm_corpustext* m_thistext;

public:
	dm_element* ElementDeclared(UnicodeString e, UnicodeString uri);
	dm_attribute* AttributeDeclared(UnicodeString e, UnicodeString,UnicodeString uri,UnicodeString,bool);
	dm_addkey* AddkeyDeclared(UnicodeString);
	bool InternalEntityDeclared(UnicodeString s);
	dm_value* ValueDeclared(dm_attribute*,UnicodeString);
	dm_addkeyvalue* AddkeyValueDeclared(dm_addkey*,UnicodeString);
	dm_namespace* NamespaceDeclared(UnicodeString);
	dm_corpustext* CorpusTextDeclared(UnicodeString);
	void MakeDSC(UnicodeString);
	void MakeGrammar(UnicodeString);
private:
	int GetNamespaceList(dm_namespace**&);
	int GetCorpusTextList(dm_corpustext**&);
	int GetElementList(dm_element**&);
	int GetInternalEntityList(dm_intent**&);
	int GetAttributesList(UnicodeString,UnicodeString,dm_attribute**&);
	int GetValueList(dm_attribute*,dm_value**&);
	int GetAddkeyValueList(dm_addkey*,dm_addkeyvalue**&);
	int GetAddkeyList(dm_addkey**&);
	UB m_ub; int* m_uu;
	int m_nw;
};
