// Indexer.cpp: implementation of the CIndexer class.
//
//////////////////////////////////////////////////////////////////////


#define _CRTDBG_MAP_ALLOC
#include "StdAfx.h"
#include "TextReader.h"
#include "Corpus.h"
#include "MemoryErr.h"
#include "Indexer.h"

#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CIndexer::CIndexer(CCorpus* c,int k,int ks):m_dict(c),m_dump(c->exppath(CCorpus::PARAM_ETCPATH,"log.txt")),
m_f("UTF-8","",&m_dump),m_clog(c->exppath(CCorpus::PARAM_ETCPATH,"corpuslog.xml")),
	m_fe("UTF-8","",&m_clog)
{
	m_if=-1;
	InitWC();
	m_bDump=c->m_bDump;
	m_lll=0;
	m_nt=0;
	m_bBrk=false;
	m_band=k; m_kstr=ks;
	m_dm.SetHeader(&c->m_r);
	m_tok=NULL;
	m_errs=0;
}

CIndexer::~CIndexer()
{
	m_fe << L("</log>") << chLF;
}

#define MAXNO 10
#define NULLVAL "0"
#define NULLNAME "0"
#define NULLURI ""

int CIndexer::Index(typ t,UnicodeString s,int l,int c,UnicodeString uri,int par)
{
	return Index(t,s,l,c,l,c,uri,par);
}

int CIndexer::Index(typ t,UnicodeString s,int l,int c,int ll,int cc,UnicodeString uri,int par)
{
	UnicodeString uri2=uri[0]?uri:NULLURI;
	switch (t) {
//	case it_content:
//		IndexWord(s,l,c);
//		break;
	case it_start:
	case it_end:{
//		indexpol0 ip=GetReader()->GetIP0(s);
//		if (ip != IP0_WTAG)
		return IndexEntry(t,UnicodeString(NULLURI),UnicodeString(NULLNAME),uri2,UnicodeString(NULLNAME),s,s,0,(UnicodeString*) NULL,l,c,ll,cc,par);
			}
		break;
	default:
		return IndexEntry(t,UnicodeString(NULLURI),UnicodeString(NULLNAME),uri2,UnicodeString(NULLNAME),s,s,0,NULL,l,c,ll,cc,par);

	}
}

void CIndexer::IndexGlobalAttr(UnicodeString n,UnicodeString s,int l,int c,UnicodeString uri,int par)
{
	m_dm.DeclareAttribute("","",n,uri,s,true);
	UnicodeString uri2=uri[0]?uri:NULLURI;
	indexpol ip=GetReader()->GetIP("",n);
	switch (ip) {
	case IP_CDATA:
		IndexEntry(it_global_attribute,uri2,n,UnicodeString(NULLURI),UnicodeString(NULLNAME),s,s,0,NULL,l,c,l,c,par);
		IndexEntryW(it_global_attribute_value,uri2,n,UnicodeString(NULLURI),UnicodeString(NULLNAME),n,n,0,NULL,l,c,l,c,s,par);

		break;
	default:
		IndexEntry(it_global_attribute,uri2,n,UnicodeString(NULLURI),UnicodeString(NULLNAME),s,s,0,NULL,l,c,l,c,par);
		break;
	}
	
}

int CIndexer::IndexStartTag(UnicodeString e,int l,int c,int ll,int cc,UnicodeString uri,bool& bC,bool& bM,int par)
{
	if (e=="bibliography" && uri==UnicodeString(XAIRA_URI)+"/bibliography") m_dm.SetInBib(true);
	unsigned ip0=GetReader()->GetIP0(e,uri); bool b;
	int pol=EIP_THIS(ip0);
	if (pol==EIP_INHERIT) b=bM;
	else b=pol==EIP_YES;
	int pol1=EIP_CONTENT(ip0);
	if (pol1 != EIP_INHERIT) 
		bC=pol1==EIP_YES;
	int pol2=EIP_CHILDREN(ip0);
	if (pol2 != EIP_INHERIT) 
		bM=pol2==EIP_YES;
	m_dm.DeclareElement(e,uri);
	if (!b) return -1;
	return Index(it_start,e,l,c,ll,cc,uri,par);
}

void CIndexer::IndexEndTag(UnicodeString e,int l,int c,int ll,int cc,UnicodeString uri,bool bM,int par)
{
	if (e=="bibliography" && uri==UnicodeString(XAIRA_URI)+UnicodeString("/bibliography")) m_dm.SetInBib(false);
	unsigned ip0=GetReader()->GetIP0(e,uri); bool b;
	int pol=EIP_THIS(ip0);
	if (pol==EIP_INHERIT) b=bM;
	else b=pol==EIP_YES;
	if (!b) return;
	m_dm.DeclareElement(e,uri);
	Index(it_end,e,l,c,ll,cc,uri,par);
}



void CIndexer::IndexAddkey(UnicodeString& n,UnicodeString& s,int l,int c,int par)
{
	m_dm.DeclareAddkey(n,s);
	IndexEntry(it_addkey,UnicodeString(NULLURI),n,UnicodeString(NULLURI),UnicodeString(NULLNAME),s,s,0,NULL,l,c,l,c,par);
}

void CIndexer::IndexNullAddkey(UnicodeString& n,UnicodeString& s)
{
	m_dm.DeclareAddkey(n,s);
}

void CIndexer::IndexInternalEntity(UnicodeString& n,UnicodeString& v)
{
	m_dm.DeclareInternalEntity(n,v);
}


void CIndexer::IndexGlobalAttrVal(UnicodeString n,UnicodeString k,int l,int c,UnicodeString uri,int par)
{
	m_dm.DeclareAttribute("","",n,uri,k,true);
	UnicodeString uri2=uri[0]?uri:NULLURI;
	IndexEntryW(it_global_attribute_value,uri2,n,UnicodeString(NULLURI),UnicodeString(NULLNAME),n,n,0,NULL,l,c,l,c,k,par);
	IndexEntry(it_global_attribute,uri2,n,UnicodeString(NULLURI),UnicodeString(NULLNAME),k,k,0,NULL,l,c,l,c,par);
}

void CIndexer::InitTax()
{
	for (int i=0;i<GetReader()->m_ntax;i++) {
		_taxonomy* t=&GetReader()->m_taxonomies[i];
		t->hastax=false;
	}
}

void CIndexer::UndefTax(int l,int c,int par)
{
	if (m_if < 2) return;
	for (int i=0;i<GetReader()->m_ntax;i++) {
		_taxonomy* t=&GetReader()->m_taxonomies[i];
		if (t->hastax) continue;
		xairo_uchar pp[10];
		_itow(t->n,pp,10);
		IndexEntryW(it_global_attribute_value,UnicodeString(XAIRA_URI),UnicodeString("catref"),UnicodeString(NULLURI),UnicodeString(NULLNAME),t->id,t->id,0,NULL,l,c,l,c,UnicodeString((const UChar*) pp),par);
		m_dm.DeclareClass(i,t->n,"","");
	}
}

void CIndexer::IndexLocalAttr(UnicodeString n,UnicodeString e,UnicodeString s,int l,int c,UnicodeString uri,int par)
{
	m_dm.DeclareAttribute(e,uri,n,"",s,false); // needs two url args?
	UnicodeString uri2=uri[0]?uri:NULLURI;
	indexpol ip=GetReader()->GetIP(e,n);
//	indexpol ip=IP_CDATA;
	switch (ip) {
	case IP_MAX: case IP_CDATA:
//		IndexEntry(it_local_attribute,UnicodeString(NULLURI),n,uri2,e,s,s,0,NULL,l,c,l,c);
//		break;
	case IP_ID: case IP_IDREF:
		IndexEntryW(it_local_attribute_value,UnicodeString(NULLURI),n,uri2,e,n,n,0,NULL,l,c,l,c,s,par);
		IndexEntry(it_local_attribute,UnicodeString(NULLURI),n,uri2,e,s,s,0,NULL,l,c,l,c,par);
		break;
	case IP_NONE:
		break;
	case IP_TAX:
		{
			UnicodeString* vals;
			int m=GetReader()->GetTaxClasses(s,vals);
			bool* bOK=new bool [m];
			for (int i=0;i<m;i++) bOK[i]=false;
			for (int i=0;i<GetReader()->m_ntax;i++) {
				int p;
				UnicodeString y;
				_taxonomy* t=&GetReader()->m_taxonomies[i];
				bool b=GetReader()->GetTaxClass(t,m,p,vals,bOK,y);
				if (b) {
					IndexEntryW(it_local_attribute_value,UnicodeString(NULLURI),n,uri2,e,t->id,t->id,0,NULL,l,c,l,c,y,par);
					xairo_uchar pp[10];
					_itow(p,pp,10);
					IndexEntryW(it_global_attribute_value,UnicodeString(XAIRA_URI),UnicodeString("catref"),UnicodeString(NULLURI),UnicodeString(NULLNAME),t->id,t->id,0,NULL,l,c,l,c,UnicodeString((const UChar*) pp),par);
					m_dm.DeclareClass(i,p,e,n);
					t->hastax=true;
				}}
				for (int i=0;i<m;i++) {
					if (bOK[i]) continue;
					xairo_uchar m[1000];
					swprintf(m,L("Unknown taxonomy class %s"),vals[i].getTerminatedBuffer());
					Error(l,c,m);
				}

//				SCWrite(t,y);
			
			mydelete [] vals;
		}

//		UnicodeString *x, *y;
//		int m=GetReader()->GetTaxClasses(s,x,y);
//		for (int i=0;i<m;i++) 
//		{
//		};		
//		mydelete [] x;
//		mydelete [] y;
//		}
		break;
	}
}



void CIndexer::IndexWord(const xairo_uchar* s,int l, int c,int ll, int cc,int par)
{
	if (m_tok->AllSpace(s)) return;
	m_dm.DeclareCharacters((const UChar*) s);
	UnicodeString ss=m_tok->IndexForm((const UChar*)s);
	UnicodeString sc=m_tok->IndexFormCase((const UChar*)s);
//	UnicodeString* cont=mynew UnicodeString [GetReader()->m_naddkeys-1];
	for (int i=1;i<GetReader()->m_naddkeys;i++) {
		_addkey* a=&GetReader()->m_addkeys[i];
		if (a->bdef)
			m_cont[i-1]=a->cache;
		else 
			switch (a->dproc) {
				case _addkey::DP_USEVAL:
					m_cont[i-1]=a->defval;
					break;
				case _addkey::DP_USEFORM:
					m_cont[i-1]=ss;
					break;
			}
	}
	IndexEntry(it_content,UnicodeString(NULLURI),UnicodeString(NULLNAME),UnicodeString(NULLURI),UnicodeString(NULLNAME),ss,sc,GetReader()->m_naddkeys-1,m_cont,l,c,ll,cc,par);
//	mydelete [] cont;
	m_nw++;
}

void CIndexer::IndexLabel(UnicodeString name,UnicodeString value,int l,int c,int par)
{
	IndexGlobalAttrVal(name,value,l,c,UnicodeString(XAIRA_URI),par);
	IndexGlobalAttr(name,value,l,c,UnicodeString(XAIRA_URI),par);
}

void CIndexer::IndexScope(int l,int c,int par)
{
	IndexGlobalAttrVal(UnicodeString("scope"),UnicodeString("0"),l,c,UnicodeString(XAIRA_URI),par);
}

static UChar s_pc_d[]={'%','d',0};

void CIndexer::Bib(int l,int c,int par,UnicodeString s)
{
	int* x=NewWC();
	m_dm.DeclareCorpusText(s,m_nw,x);
	int nb=GetReader()->m_nbib;
	if (m_nt > 0 && m_nt <= nb) {
		int b1=GetReader()->m_bib[m_nt-1].start;
		int b2=GetReader()->m_bib[m_nt-1].length;
		xairo_uchar x[10];
		swprintf(x,s_pc_d,b1);
		IndexGlobalAttrVal(UnicodeString("BIBSTART"),UnicodeString((const UChar*) x),l,c,UnicodeString(XAIRA_URI),par);
		swprintf(x,s_pc_d,b2);
		IndexGlobalAttrVal(UnicodeString("BIBLENGTH"),UnicodeString((const UChar*) x),l,c,UnicodeString(XAIRA_URI),par);
	}
	m_nt++;
}

#define DUMPRAWINDEX true


int CIndexer::IndexEntry(typ t,UnicodeString uri,UnicodeString n,UnicodeString urie,UnicodeString e,UnicodeString s,UnicodeString s1,int nco,UnicodeString* co,int l,int c,int ll, int cc,int par)
{
	return IndexEntryW(t,uri,n,urie,e,s,s1,nco,co,l,c,ll,cc,m_nw,par);
}

//void CIndexer::IndexEntry(typ t,const XMLCh* uri,const XMLCh* n,const XMLCh* urie,const XMLCh* e,const XMLCh* s,const XMLCh* s1,int o)
//{
//	IndexEntryW(t,uri,n,urie,e,s,s1,o,m_nw);
//}

static int oldo;

void CIndexer::Premap(int l,int c)
{
	m_reader->Premap(l,c);
}

void CIndexer::Premap2(int l,int c)
{
	m_reader->Premap2(l,c);
}

int CIndexer::IndexEntryW(typ t,UnicodeString uri,UnicodeString n,UnicodeString urie,UnicodeString e,UnicodeString s,UnicodeString s1,int nco, UnicodeString* co,int l,int c,int ll, int cc,UnicodeString w,int par)
{
/*	
Only do this if the user wants defaulted line number markup
  while (l>0 && m_l0<l && !m_bLine) {
		m_bLine=true;
		Index(it_start,LINE,m_l0+1,1,XAIRA_URI);
		int k=m_reader->GetLineOffset(m_l0);
		xairo_uchar x[MAXNO];
		_itow(k,x,10);
		IndexGlobalAttr(OFFSET,x,m_l0+1,1,XAIRA_URI);
		Index(it_end,LINE,m_l0+1,1,XAIRA_URI);
		m_l0++;
		m_bLine=false;
	}
	*/
	TryLineLabel(l,par);
	int o=m_reader->Offset(l,c);
	int oo=m_reader->Offset(ll,cc);

//	assert(o >= oldo); 
	oldo=o;
	Break(m_if,o);
	int x=m_dict.GetSeq();
	m_dict.Index(t,uri,n,urie,e,s,s1,nco,co,m_if,w,o,oo-o,par);
	if (m_bDump) {
		m_f << (const xairo_uchar*) Lname(t).getTerminatedBuffer() 
			<< chSpace << (const xairo_uchar*) uri.getTerminatedBuffer() 
			<< chSpace << (const xairo_uchar*) n.getTerminatedBuffer() 
			<< chSpace << (const xairo_uchar*) urie.getTerminatedBuffer() 
			<< chSpace << (const xairo_uchar*) e.getTerminatedBuffer() 
			<< chSpace << (const xairo_uchar*) s.getTerminatedBuffer();
		for (int i=0;i<nco;i++) {
			m_f << chSpace << (const xairo_uchar*) co[i].getTerminatedBuffer();
		}
		m_f << chSpace << (const xairo_uchar*) IntToWide(m_if).getTerminatedBuffer();
		m_f << chSpace << (const xairo_uchar*) w.getTerminatedBuffer();
		m_f << chSpace << (const xairo_uchar*) IntToWide(o).getTerminatedBuffer();
		m_f << chLF;
	}
	return x;

}



int CIndexer::IndexEntryW(typ t,UnicodeString uri,UnicodeString n,UnicodeString urie,UnicodeString e,UnicodeString s,UnicodeString ss,int nco, UnicodeString* co,int l,int c,int ll, int cc, int w,int par)
{
/*	
Only do this if the user wants defaulted line number markup
  while (l>0 && m_l0<l && !m_bLine) {
		m_bLine=true;
		Index(it_start,LINE,m_l0+1,1,XAIRA_URI);
		int k=m_reader->GetLineOffset(m_l0);
		xairo_uchar x[MAXNO];
		_itow(k,x,10);
		IndexGlobalAttr(OFFSET,x,m_l0+1,1,XAIRA_URI);
		Index(it_end,LINE,m_l0+1,1,XAIRA_URI);
		m_l0++;
		m_bLine=false;
	}
	*/
	TryLineLabel(l,par);
	int o=m_reader->Offset(l,c);
	int oo=m_reader->Offset(ll,cc);
//	assert(o >= oldo); 
	oldo=o;
	Break(m_if,o);
	int x=m_dict.GetSeq();
	m_dict.Index(t,uri,n,urie,e,s,ss,nco,co,m_if,w,o,oo-o,par);
	if (m_bDump) {
		m_f << (const xairo_uchar*) Lname(t).getTerminatedBuffer() << 
			chSpace << (const xairo_uchar*) uri.getTerminatedBuffer() 
			<< chSpace << chSpace << (const xairo_uchar*) n.getTerminatedBuffer() << 
			chSpace << chSpace << (const xairo_uchar*) urie.getTerminatedBuffer() << 
			chSpace << chSpace << (const xairo_uchar*) e.getTerminatedBuffer() << 
			chSpace << chSpace << (const xairo_uchar*) s.getTerminatedBuffer() <<
			chSpace << chSpace << (const xairo_uchar*) ss.getTerminatedBuffer();
		for (int i=0;i<nco;i++) {
			m_f << chSpace << chSpace << (const xairo_uchar*) co[i].getTerminatedBuffer();
		}
		m_f << chSpace << chSpace << (const xairo_uchar*) IntToWide(m_if).getTerminatedBuffer();
		m_f << chSpace << chSpace << (const xairo_uchar*) IntToWide(w).getTerminatedBuffer();
		m_f << chSpace << chSpace << (const xairo_uchar*) IntToWide(o).getTerminatedBuffer();
		m_f << chLF;
	}
	return x;

}
//void CIndexer::IndexEntryW(typ t,const XMLCh* uri,const XMLCh* n,const XMLCh* urie,const XMLCh* e,const XMLCh* s,const XMLCh* ss,int o,int w)
//{
/*	
Only do this if the user wants defaulted line number markup
  while (l>0 && m_l0<l && !m_bLine) {
		m_bLine=true;
		Index(it_start,LINE,m_l0+1,1,XAIRA_URI);
		int k=m_reader->GetLineOffset(m_l0);
		xairo_uchar x[MAXNO];
		_itow(k,x,10);
		IndexGlobalAttr(OFFSET,x,m_l0+1,1,XAIRA_URI);
		Index(it_end,LINE,m_l0+1,1,XAIRA_URI);
		m_l0++;
		m_bLine=false;
	}
	*/
//	m_dict.Index(t,uri,n,urie,e,s,ss,m_if,w,o);
//	if (DUMPRAWINDEX) {
//		m_f << Lname(t) << chSpace << uri << chSpace << n << chSpace << urie << chSpace << e << chSpace << s <<chSpace << ss;
//		m_f << chSpace << IntToWide(m_if);
//		m_f << chSpace << IntToWide(w);
//		m_f << chSpace << IntToWide(o);
//		m_f << chLF;
//	}
//
//}



void CIndexer::Error(const SAXParseException& e)
{
	xairo_uchar buffer[1000];
	const xairo_uchar* m=e.getMessage();
	swprintf(buffer,L("<loc sysid=\"%s\" line=\"%d\" col=\"%d\"/><msg>%s</msg>"),
		e.getSystemId(),e.getLineNumber(),e.getColumnNumber(),e.getMessage());
	Log(L("parse-error"),buffer);
}

static UChar efmt[]={'<','l','o','c',' ','s','y','s','i','d','=','\"','%','s','\"',' ','l','i','n','e',
'=','\"','%','d','\"',' ','c','o','l','=','\"','%','d','\"','/','>','<','m','s','g','>','%','s','<','/','m','s','g','>',0};
static UChar hefmt[]={'<','l','o','c',' ','s','y','s','i','d','=','\"','h','e','a','d','e','r','\"',
'/','>','<','m','s','g','>','%','s','<','/','m','s','g','>',0};
static UChar s_error[]={'e','r','r','o','r',0};


void CIndexer::Error(int l,int c,const xairo_uchar* msg)
{
	xairo_uchar buffer[1000];
	swprintf(buffer,efmt,
		m_name.getTerminatedBuffer(),l,c,msg);
	Log(s_error,buffer);
	m_errs++;
}

void CIndexer::HeaderError(const xairo_uchar* msg)
{
	xairo_uchar buffer[1000];
	swprintf(buffer,hefmt,msg);
	Log(s_error,buffer);
	m_errs++;
}


void CIndexer::Note(XMLCh* e)
{
	Log(L("note"),e);
}

void CIndexer::Log(const XMLCh* type,const XMLCh* e,...)
{
	va_list list; 
	va_start(list,e);
	time_t t;
	time(&t);
	const xairo_uchar* b=_wctime(&t);
	xairo_uchar buffer[1000];
	vswprintf(buffer,e,list);
	va_end(list);
	m_fe << L("<entry type=\"") << type << L("\" time=\"") << b << L("\">");
	m_fe << buffer;
	m_fe << L("</entry>\n");
	if (wcscmp(type,L("process"))==0) {
		char* t=mynew char[wcslen(buffer)+1];
		wcstombs(t,buffer,wcslen(buffer)+1);
		printf(t);
		printf("\n");
		delete [] t;
	}
}


UnicodeString CIndexer::Lname(typ t)
{
	switch (t) {
	case it_content:
		return "CONTENT";
	case it_start:
		return "START";
	case it_end:
		return "END";
	case it_global_attribute:
		return "GATT";
	case it_local_attribute:
		return "LATT";
	case it_pi:
		return "PI";
	case it_global_attribute_value:
		return "GATTV";
	case it_local_attribute_value:
		return "LATTV";
	case it_content_attribute:
		return "CATT";
	case it_enddoc:
		return "ENDDOC";
	case it_addkey:
		return "AK";
	case it_spm:
		return "SPM";
	case it_epm:
		return "EPM";
	}
	assert(false);
	return "ERR";
}

void CIndexer::IndexPI(UnicodeString n, UnicodeString s,int l, int c,int par)
{
	IndexEntry(it_pi,UnicodeString(NULLURI),n,UnicodeString(NULLURI),UnicodeString(NULLNAME),s,s,0,NULL,l,c,l,c,par);
}

void CIndexer::IndexSPM(UnicodeString prefix, UnicodeString uri,int l, int c,int par)
{
	IndexEntryW(it_spm,UnicodeString(NULLURI),UnicodeString(NULLNAME),UnicodeString(NULLURI),UnicodeString(NULLNAME),prefix,prefix,0,NULL,l,c,l,c,uri,par);
}

void CIndexer::IndexEPM(UnicodeString prefix,int l, int c,int par)
{
	IndexEntry(it_epm,UnicodeString(NULLURI),UnicodeString(NULLNAME),UnicodeString(NULLURI),UnicodeString(NULLNAME),prefix,prefix,0,NULL,l,c,l,c,par);
}


void CIndexer::StartCorpus(UnicodeString n,int k,_lang* l)
{
	m_dict.Open(this,m_band,m_kstr,l?l->locale:"");
	Index(it_start,UnicodeString(CORPUS),0,1,UnicodeString(XAIRA_URI),0);
	IndexGlobalAttr(UnicodeString(NAME),n,0,1,UnicodeString(XAIRA_URI),0);
	IndexGlobalAttr(UnicodeString(TEXTS),IntToWide(k),0,1,UnicodeString(XAIRA_URI),0);
	if (l) {IndexGlobalAttrVal(UnicodeString("locale"),UnicodeString(l->locale),0,1,UnicodeString(XAIRA_URI),0);
	
	if (l->bUserTokenRules) 
		IndexGlobalAttrVal(UnicodeString("wbrules"),l->rules,0,1,UnicodeString(XAIRA_URI),0);
	} else IndexGlobalAttrVal(UnicodeString("locale"),UnicodeString("#def"),0,1,UnicodeString(XAIRA_URI),0);

	m_bDoc=false;
	m_cont=mynew UnicodeString[GetReader()->m_naddkeys-1];
	
}

void CIndexer::EndCorpus()
{
//	NewFile();
	oldo=0;
	int* x=NewWC();
	Index(it_end,UnicodeString(CORPUS),0,1,UnicodeString(XAIRA_URI),0);
	m_dm.DeclareCorpusText("",m_nw,x);
//	delete [] m_cont;
}

void CIndexer::EndLog()
{	
}


void CIndexer::Push()
{
	_ctxt* c=mynew _ctxt;
	c->reader=m_reader;
	c->iif=m_if;
	c->nw=m_nw;
	c->line=m_l0;
	c->lll=m_lll;
//	c->col=m_c0;
	c->oldo=oldo;
	c->next=m_stack;
	m_stack=c;
	m_if=m_dict.m_corpus->m_nif++;
//	m_nw=0;
	m_l0=0;
	oldo=0;
	m_lll=0;
}

void CIndexer::Pop()
{
	_ctxt* c=m_stack;
	m_stack=c->next;
	mydelete m_reader;
	m_reader=c->reader;
	m_if=c->iif;
//	m_nw=c->nw;
	m_l0=c->line;
	m_lll=c->lll;
//	m_c0=c->col;
	oldo=c->oldo;
}

void CIndexer::NewFile(int i)
{
	m_if=m_dict.m_corpus->m_nif++;
//	m_nw=0;
	m_l0=0;
	m_bLine=false;
	m_lll=0;
//	m_errs=0;
	oldo=0;
	for (int j=0;j<GetReader()->m_naddkeys;j++) {
		GetReader()->m_addkeys[j].bdef=false;
	}
	for (int j=0;j<GetReader()->m_nlabels;j++) {
		GetReader()->m_labels[j].bdef=false;
	}
}

UnicodeString IntToWide(int i)
{
	UErrorCode e=U_ZERO_ERROR;
	UnicodeString s;
	DecimalFormat df(e);
	df.format((int32_t) i,s);
	return s;
}

UnicodeString FloatToWide(double f)
{
	UErrorCode e=U_ZERO_ERROR;
	UnicodeString s;
	DecimalFormat df(e);
	df.format(f,s);
	return s;
}

void CIndexer::SkipText(UnicodeString name)
{
	m_name=name;
	Log(L("process"),L("Skipped %s (see log)"),name.getTerminatedBuffer()); 
	dummyDocument(name,0);
}

void CIndexer::StartText(UnicodeString name,UnicodeString relpath)
{
	m_name=name;
	m_relpath=relpath;
	Log(L("process"),L("Text %s"),name.getTerminatedBuffer());
}

void CIndexer::EndText(UnicodeString name,int n)
{
	if (n) Log(L("process"),L("%d errors in %s"),n,name.getTerminatedBuffer());
	m_dict.Report();
//	int m;
//	if (m_dict.Balance(m)) Log(L"process",L"Balance - max was %d",m);
}

void CIndexer::Phase2()
{
	Log(L("process"),L("Phase 2"));
	m_dict.Close();
	Stats();
	m_dict.Save(m_nt);
}

CSaraHeaderReader* CIndexer::GetReader()
{
	return &m_dict.m_corpus->m_r;
}


void CIndexer::TryLineLabel(int ln,int par)
{
	int n=GetReader()->m_nlabels;
	if (n<2) return;
	_lab* l=&GetReader()->m_labels[n-1];
	if (!l->isSpec) return;
	if (l->e != SPE_LINE) return;
	if (ln <= m_lll) return;
	XMLCh no[MAXNO];
	swprintf(no,s_pc_d,ln);
	m_lll=ln;
	IndexLabel(UnicodeString("unit"),UnicodeString((const UChar*) no),ln,1,par);
	IndexScope(ln,1,par);
}



void CIndexer::SCWrite(_taxonomy* tax,UnicodeString val)
{
	UnicodeString path=m_dict.m_corpus->GetSCPath(tax->id,"sc");
	if (path[0]==0) return;
	FILE* f=_wfopen((const xairo_uchar*) path.getTerminatedBuffer(),L("a"));
	int i=GetReader()->GetSCPart(tax,val);
	fwprintf(f,L("%d\n"),i);
	fclose(f);
}

void CIndexer::dummyDocument(UnicodeString name,int par)
{
	Index(it_start,UnicodeString(FILEN),0,1,UnicodeString(XAIRA_URI),par);
	UnicodeString e=m_reader->m_enc;
	IndexGlobalAttrVal(UnicodeString(ENCODING),e,0,1,UnicodeString(XAIRA_URI),par);
	IndexGlobalAttrVal(UnicodeString(NAME),name,0,1,UnicodeString(XAIRA_URI),par);
	Index(it_end,UnicodeString(FILEN),0,1,UnicodeString(XAIRA_URI),par);
}

// note that what follows is Windows specific
#ifdef __unix__
// commented out because not in FreeBSD
//#include <sys/resource.h>
//
void CIndexer::Stats()
{
//	struct rlimit rl; struct rusage ru;
//	Log(L("note"),L("About to dump global memory statistics."));
//	getrlimit(RLIMIT_AS,&rl);
//	getrusage(RUSAGE_SELF,&ru);
//	Log(L("stat"),L("<stat name=\"memlimsoft\">%d</stat>"),rl.rlim_cur);
//	Log(L("stat"),L("<stat name=\"memlimhard\">%d</stat>"),rl.rlim_max);
//	Log(L("stat"),L("<stat name=\"idrss\">%d</stat>"),ru.ru_idrss);
//	Log(L("note"),L("Global memory statistics dump complete."));
//
}
#elif defined (_WIN32)

/*
#include "windows.h"
*/
void CIndexer::Stats()
{
/*	if (!m_bMem) return;
	MEMORYSTATUSEX e;
	e.dwLength=sizeof e;
	GlobalMemoryStatusEx(&e);
	Log(L"note",L"About to dump global memory statistics.");
	Log(L"stat",L"<stat name=\"gtp\">%I64d</stat>",e.ullTotalPhys);
	Log(L"stat",L"<stat name=\"gap\">%I64d</stat>",e.ullAvailPhys);
	Log(L"stat",L"<stat name=\"gtpf\">%I64d</stat>",e.ullTotalPageFile);
	Log(L"stat",L"<stat name=\"gapf\">%I64d</stat>",e.ullAvailPageFile);
	Log(L"stat",L"<stat name=\"gtv\">%I64d</stat>",e.ullTotalVirtual);
	Log(L"stat",L"<stat name=\"gav\">%I64d</stat>",e.ullAvailVirtual);
	Log(L"note",L"Global memory statistics dump complete.");
*/
}

#else

void CIndexer::Stats()
{
}

#endif

void CIndexer::SetBreak(char* s)
{
	if (sscanf(s,"%d/%d",&m_brkfi,&m_brkloc)==2)
		m_bBrk=true;
}

void CIndexer::Break(int f,int o)
{
#ifdef _DEBUG
	if (!m_bBrk) return;
	if (f != m_brkfi || o != m_brkloc) return;
	__asm { int 3 };
#endif
}

void CIndexer::DSC() 
{
	UnicodeString path=m_dict.m_corpus->exppath(CCorpus::PARAM_INDEXPATH,"dsc.txt");
	m_dm.MakeDSC(path);
	UnicodeString pathg=m_dict.m_corpus->exppath(CCorpus::PARAM_INDEXPATH,"xgrammar.xml");
	m_dm.MakeGrammar(pathg);
}

void CIndexer::SetLang(const char* l)
{
	m_lang=l;
	if (m_tok) delete m_tok;
	m_tok=new CTokeniser(l);
}


void CIndexer::Regions(int i,UnicodeString val)
{
	CSaraHeaderReader* h=GetReader();
	int n=h->m_nreg;
	for (int ii=0;ii<n;ii++) {
		_region* r=&h->m_reg[ii];
		if (r->nak != i) continue;
		bool bOK=false;
		for (int j=0;j<r->n;j++) {
			if (r->a[j]==val) bOK=true;
		}
		if (bOK==m_rmask[ii]) continue;
		if (bOK)
			m_w1[ii]=m_nw;
		else	
			m_w2[ii]+=m_nw-m_w1[ii];
		m_rmask[ii]=bOK;
	}
}

void CIndexer::InitWC()
{
	m_nw=0;
	CSaraHeaderReader* h=GetReader();
	m_w1=new int [h->m_nreg];
	m_w2=new int [h->m_nreg];
	m_rmask=new bool [h->m_nreg];
	for (int i=0;i<h->m_nreg;i++) {
		m_w1[i]=m_w2[i]=0;
		m_rmask[i]=false;
	}
}

int* CIndexer::NewWC()
{
	CSaraHeaderReader* h=GetReader();
	int* x=new int [h->m_nreg];
	for (int i=0;i<h->m_nreg;i++) {
		x[i]=m_w2[i];
		m_w1[i]=m_w2[i]=0;
		m_rmask[i]=false;
	}
	return x;
}



