// TextReader.cpp: implementation of the CTextReader class.
//
//////////////////////////////////////////////////////////////////////

#define _CRTDBG_MAP_ALLOC
#include "StdAfx.h"

#include "Corpus.h"
#include "Indexer.h"
#include "TextReader.h"


#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTextReader::CTextReader(UnicodeString p,CIndexer* ind)
{
	m_path=p;
	m_i=m_j=0;
	m_lines=NULL;
	m_bold=false;
	m_cs=NULL;
	m_linels=NULL;
	m_pl=m_pc=-1;
	m_lineo=NULL;
	m_ind=ind;
}

CTextReader::~CTextReader()
{
	Free();
	if (m_lines) mydelete m_lines;
}

static UChar s_rb[]={'r','b',0};

bool CTextReader::Read(int& nb)
{
	FILE* f=iopen(m_path,s_rb);
	if (f==NULL) 
	{
		m_ind->Log(L("error"),L("Cannot open file %s"),m_path.getTerminatedBuffer());
		return false;
	}
	if (fseek(f,0,SEEK_END) != 0) 
	{
		m_ind->Log(L("error"),L("Cannot open file %s"),m_path.getTerminatedBuffer());
		return false;
	}
	int k=ftell(f);
	fseek(f,0,SEEK_SET);
	char* input=mynew char [k+1];
	char* i0=input;
	int n=fread(input,1,k,f);
	if (n<k) 
	{
		m_ind->Log(L("error"),L("Error reading file %s"),m_path.getTerminatedBuffer());
		return false;
	}
	input[n]=0;
	fclose(f);
// Now puzzle out the encoding. See XML App F.1
	m_p=GuessEncoding(input,n);
	input+=m_p;
	k-=m_p;
#ifdef USE_XERCES_TRANSCODERS
	bool b=XercesTranscode(input,k);
#else
	bool b=ICUTranscode(input,k);
#endif
	mydelete [] i0;
	if (!b || m_b < (unsigned) k) {
		Free();
		return false;
	}
	LineLengths();
	nb=m_kk;
	return true;
}

bool CTextReader::XercesTranscode(const char* input,XMLSize_t k)
{
	XMLTransService *s=XMLPlatformUtils::fgTransService;
	XMLTransService::Codes c=XMLTransService::Ok; 
	XMLTranscoder* t=s->makeNewTranscoderFor((const xairo_uchar*) m_enc.getTerminatedBuffer(),c,k,XMLPlatformUtils::fgMemoryManager);
	if (c!=XMLTransService::Ok) {
//		mydelete i0;
		m_ind->Log(L("error"),L("Bad encoding %s"),m_enc.getTerminatedBuffer());
		return false;
	}
	m_d=""; 
	m_cs=mynew unsigned char [k];
	try {
		m_kk=t->transcodeFrom((unsigned char*) input,k,m_d.getBuffer(k),k,m_b,m_cs);
		m_d.releaseBuffer(k);
	}
	catch (TranscodingException& err)
	{
		m_ind->Log(L("error"),L("Encoding error %s at line %d"),err.getMessage(),err.getSrcLine());
	}
	return true;
}

bool CTextReader::ICUTranscode(const char* input,int k)
{
	UErrorCode err=U_ZERO_ERROR;
	UConverter* u=ucnv_openU(m_enc.getTerminatedBuffer(),&err);
	if (err != U_ZERO_ERROR) {
		m_ind->Log(L("error"),L("Bad encoding %s"),m_enc.getTerminatedBuffer());
		return false;
	}
	UChar* target=m_d.getBuffer(k);
	UChar* t0=target;
	int* os=mynew int [k+1];
	const char* i0=input;
	ucnv_toUnicode(u,&target,target+k,&input,input+k,os,TRUE,&err);
	m_d.releaseBuffer(k);
	m_cs=mynew unsigned char [k];
	m_kk=target-t0;
	m_b=input-i0;
	os[m_kk]=m_b;
	for (int i=0;i<m_kk;i++) {
		m_cs[i]=os[i+1]-os[i];
	}
	mydelete [] os;
	return true;
}


InputSource* CTextReader::GetInputSource()
{
	InputSource* is=mynew MemBufInputSource((const unsigned char*) m_d.getTerminatedBuffer(),m_kk*2,L("C:\\sysid.xml"),false);
	is->setSystemId((const xairo_uchar*) m_path.getTerminatedBuffer());
	is->setEncoding(L("UTF-16"));
	return is;
}

void CTextReader::LineLengths()
{
	m_nlines=0;
	int i;
	for (i=0;i<m_kk;i++) {
		if (m_d[i]==10) m_nlines++;
		if (m_d[i]==13 && m_d[i+1]==10) {i++; m_nlines++;}
		if (m_d[i]==13) m_nlines++;
	}
	m_nlines++;
	m_linels=mynew int [m_nlines];
	m_lineo=mynew int [m_nlines];
	int c=0; int l=0; int x=0;
	unsigned char* cs=m_cs;
	for (i=0;i<m_kk;i++) {
		if (m_d[i]==10) {
			m_lineo[l]=x+*cs;
			m_linels[l++]=c;
			c=0;
		} else if (m_d[i]==13 && m_d[i+1]==10) {
			i++;
			x+=*cs;
			cs++;
			m_lineo[l]=x+*cs;
			m_linels[l++]=c;
			c=0;
		} else if (m_d[i]==13) {
			m_lineo[l]=x+*cs;
			m_linels[l++]=c;
			c=0;
		} else
			c++;
		x+=*cs;
		cs++;
	}
	m_linels[l]=c;
	m_lineo[l]=x;
	m_pl3=-1;
}



void CTextReader::Free()
{
	if (m_cs) mydelete[] m_cs;
	if (m_linels) mydelete[] m_linels;
	if (m_lineo) mydelete[] m_lineo;
	m_cs=NULL;
	m_linels=NULL;
	m_lineo=NULL;
}

void CTextReader::Premap(int l,int c)
{
	m_po=Offset(l,c);
	m_pl=l;
	m_pc=c;
}

void CTextReader::Premap2(int l,int c)
{
	m_po2=Offset(l,c);
	m_pl2=l;
	m_pc2=c;
}

int CTextReader::Offset(int li,int co)
{
	int l1=1; int c1=1; int k=0;  int nmiss=0;
	if (li==0) return 0;
	if (li==m_pl && co==m_pc) return m_po;
	if (li==m_pl2 && co==m_pc2) return m_po2;
	if (co==1) return li==1?0:m_lineo[li-2]+m_p;
	unsigned char* cs=m_cs; int dind=0;
	if (m_bold && (m_oldli < li || (m_oldli==li && m_oldco <= co))) {
		l1=m_oldli; c1=m_oldco; k=m_oldk;
		dind=m_oldd; cs=m_cs+m_oldd; }
	else {
		if (m_bold) {
			nmiss++;
		}
	}
	while (l1 < li || (l1==li && c1 < co)) {
		if (dind >= m_kk) break;
		xairo_uchar w=m_d[dind];
		if (w=='\n') {
			l1++;
			c1=1;
		} else if (w=='\r' && m_d[dind+1]=='\n') {
			l1++; c1=1; dind++; k+=*cs; cs++;
		} else if (w=='\r') {
			l1++; c1=1;}
		else
			c1++;
//		assert(*cs==1);
		k+=*cs;
		dind++; cs++;
	}
	m_oldd=dind;
	m_oldk=k;
	m_oldli=li;
	m_oldco=co;
	m_bold=true;
	return k+m_p;
}




long CTextReader::GuessEncoding(char* s,int ns)
{
	UErrorCode e=U_ZERO_ERROR; int32_t l=0;
	const char* c=ucnv_detectUnicodeSignature(s,ns,&l,&e);
	if (l > 0) {
		m_enc=c;
		return l;
	}
	m_enc=(const UChar*) L("UTF-8");
	if (s[0]==0xef && s[1]==0xbb && s[2]==0xbf) return 3; // not sure this is legal in XML
// Must be 8 bit then. OK it could be EBCDIC, sigh
	if (s[0]=='<' && s[1]=='?' && s[2]=='x' && s[3]=='m') {
		char* x=strchr(s,'>');
		if (x==NULL) return 0;
		char* y=strstr(s,"encoding=\"");
		if (y != NULL) y+=10;
		if (y==NULL || y > x) return 0;
		m_enc="";
		for (int i=0;i<MAXENCODING;i++) {
			if (y+i >= x || *(y+i)=='\"') break;
			m_enc+=(UChar) *(y+i);
		}
	}
	return 0;
}

#define INCR 100


void CTextReader::RecordLine(int off)
{
	if (m_i >= m_j) {
		int* old=m_lines;
		m_lines=mynew int[m_j+INCR];
		if (old != NULL) {
			for (int i=0;i<m_j;i++)
				m_lines[i]=old[i];
			mydelete old;
		}
		m_j+=INCR;
	}
	m_lines[m_i++]=off;
	return;
}






int CTextReader::GetLineOffset(int i)
{
	if (i<m_i) return m_lines[i];
	return -1;
}


void CTextReader::DumpAll(const char* standard)
{
	int n=ucnv_countAvailable();
	for (int i=0;i<n;i++) {
		const char* n=ucnv_getAvailableName(i);
		const char* nn=NULL;
		UErrorCode e=U_ZERO_ERROR;
		if (standard) nn=ucnv_getStandardName(n,standard,&e);
		printf("%s\n",nn?nn:n);
	}
}