// WordList.cpp: implementation of the CWordList class.
//
//////////////////////////////////////////////////////////////////////

#define _CRTDBG_MAP_ALLOC
#include "StdAfx.h"

#include "WordList.h"
#include "Corpus.h"

#undef THIS_FILE
static char THIS_FILE[] = __FILE__;

#if defined(_DEBUG) && !defined(_TOOLS)
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



CWordList::CWordList()
{
	m_init=100000;
	m_expand=100000;
	m_lhc=0;
	m_places=6;
	m_shift=5;
	m_hashsize=m_size=m_init;	
	m_idx=new int[m_hashsize];
	m_tab=new xara_wl[m_size];
	m_nchars=0; int i;
	for (i=0;i<m_hashsize;i++) {
		m_idx[i]=-1;
	}
	for (i=0;i<m_size;i++) {
		m_tab[i].dicptr=-1;
	}
	m_next=0;
	m_bExp=false;
	m_nmaxword=0;
}

CWordList::~CWordList()
{
	delete m_idx;
	for (int i=0;i<m_next;i++)
		delete m_tab[i].word;
	delete m_tab;
}

int CWordList::Add(const xairo_uchar* w)
{
	if (!m_bExp && m_next >= m_size) Expand();
	int k=hash(w);
	int* n=&m_idx[k];
	int hc=0;
	while (*n != -1) {
		if (wcscmp(w,m_tab[*n].word)==0) {
			assert(!m_bExp);
			return *n;
		}
		n=&m_tab[*n].next;
		hc++;
	}
	*n=m_next++;
	m_tab[*n].word=_wcsdup(w);
	m_nchars+=wcslen(m_tab[*n].word)+1;
	m_tab[*n].next=-1;
	if (hc > m_lhc) m_lhc=hc;
	if ((int) wcslen(w) > m_nmaxword) {
		m_nmaxword=wcslen(w);
		m_maxword=*n;
	}
	return *n;
}

int CWordList::ExpAdd(const xairo_uchar* w)
{
	int k=hash(w);
	int* n=&m_idx[k];
	while (*n != -1) {
		n=&m_tab[*n].next;
	}
	*n=m_next++;
	m_tab[*n].word=w;
	m_tab[*n].next=-1;
	return *n;
}

void CWordList::Dump()
{
	FILE* f=fopen("wld.txt","w");
	for (int i=0;i<m_next;i++) {
		fwrite(m_tab[i].word,2,wcslen(m_tab[i].word),f);
		fwrite("\n\0",2,1,f);
	}
	fclose(f);
}



int CWordList::hash(const xairo_uchar* w)
{
	unsigned int k=0;
//	int n=w.length();
//	if (n > m_places) n=m_places;
//	for (int i=0;i<n;i++) {
	for (int i=0;i<m_places;i++) {
		if (w[i]==0) break; // new
//		k <<= m_shift;
//		k+=w[i];
		k+=(w[i] << m_shift*i);
	}
	return k % m_hashsize;
}

const xairo_uchar* CWordList::Get(int i)
{
	assert(i<m_next);
	return m_tab[i].word;
}

void CWordList::Expand()
{
	m_bExp=true;
	int n=m_size;
	delete [] m_idx;
	xara_wl* oldw=m_tab;
	m_size+=m_expand;
	m_hashsize+=m_expand;
	m_idx=new int[m_hashsize];
	m_tab=new xara_wl[m_size]; int i;
	for (i=0;i<m_hashsize;i++) {
		m_idx[i]=-1;
	}
	for (i=0;i<m_size;i++) {
		m_tab[i].dicptr=-1; // eek
	}
	m_next=0;
//	m_nchars=0;
	for (i=0;i<n;i++) {
		int k=ExpAdd(oldw[i].word);
//		delete [] oldw[i].word;
		m_tab[k].dicptr=oldw[i].dicptr;
	}
	delete [] oldw;
	m_bExp=false;
}
	
void CWordList::Save(UnicodeString path,int nc,xara_wl_header_addkey * k,int nt,int nls,xara_wl_header_tax* kk,int lemma)
{
	FILE* f=iopen(path,(const UChar*) L("wb"));
	int n=m_next;
	xara_wl_header h;
	h.sig=DBSIG;
	h.version=DBVER;
	h.size=n;
	h.hashsize=m_hashsize;
	h.places=m_places;
	h.shift=m_shift;
	h.cont=nc;
	h.lemma=lemma;
	h.ntax=nt;
	h.nls=nls;
	fwrite(&h,sizeof (h),1,f);
	fwrite(k,sizeof(xara_wl_header_addkey),nc,f);
	fwrite(kk,sizeof(xara_wl_header_tax),nt,f);
	fwrite(m_idx,sizeof(int),m_hashsize,f); int i;
	for (i=0;i<n;i++) {
		xara_wl_file fl;
		fl.len=wcslen(m_tab[i].word);
		fl.next=m_tab[i].next;
		fl.dicptr=m_tab[i].dicptr;
		fwrite(&fl,sizeof(fl),1,f);
	}
	for (i=0;i<n;i++) {
		fwrite(m_tab[i].word,2,wcslen(m_tab[i].word),f);
	}
	fclose(f);	
}
