//#include "StdAfx.h"
#ifndef _WIN32


#include "tinylib.h"

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
 
#ifndef xairo_uchar
typedef unsigned short xairo_uchar;
#endif


const xairo_uchar* L(const char* x)
{
	xairo_uchar* w=new xairo_uchar [strlen(x)+1];
	mbstowcs(w,x,strlen(x));
	return w;
}

xairo_uchar* wcscat (xairo_uchar *dest, const xairo_uchar *src)
{
  register xairo_uchar *s1 = dest;
  register const xairo_uchar *s2 = src;
  xairo_uchar c;

  /* Find the end of the string.  */
  do
    c = *s1++;
  while (c != L'\0');

  /* Make S1 point before the next character, so we can increment
     it while memory is read (wins on pipelined cpus).	*/
  s1 -= 2;

  do
    {
      c = *s2++;
      *++s1 = c;
    }
  while (c != L'\0');

  return dest;
}

xairo_uchar *wcscpy (xairo_uchar* dest,const xairo_uchar*  src)
{
  xairo_uchar c;
  xairo_uchar *wcp;

      wcp = dest;

      do
	{
	  c = *src++;
	  *wcp++ = c;
	}
      while (c != L'\0');

  return dest;
}

xairo_uchar *wcsncpy (xairo_uchar* dest, const xairo_uchar* src, size_t n)
{
  xairo_uchar c;
  xairo_uchar *const s = dest;
  if (n==0) return s;
  do
    {
      c = *src++;
      *dest++ = c;
      if (--n == 0)
	return s;
    }
  while (c != L'\0');

 zero_fill:
  do
    *dest++ = L'\0';
  while (--n > 0);

  return s;
}

xairo_uchar *wcsncat (xairo_uchar* dest, const xairo_uchar* src, size_t n)
{
  xairo_uchar c;
  xairo_uchar * const s = dest;
	if (n==0) return s;
  /* Find the end of DEST.  */
  do
    c = *dest++;
  while (c != L'\0');
  dest--;

  while (n > 0)
    {
      c = *src++;
      *dest++ = c;
      if (c == L'\0')
	return s;
      n--;
    }

  if (c != L'\0')
    *dest++ = L'\0';

  return s;
}

int wcsncmp(const xairo_uchar* a,const xairo_uchar* b,size_t n)
{
	for (int i=0;i<n;i++) {
		if (a[i] != b[i]) return a[i]-b[i];
		if (a[i]==0) return 0;
	}
	return 0;
}


int wcslen (const xairo_uchar* s)
{
  int len = 0;

  while (s[len] != L'\0')
    {
      if (s[++len] == L'\0')
	return len;
      if (s[++len] == L'\0')
	return len;
      if (s[++len] == L'\0')
	return len;
      ++len;
    }

  return len;
}

xairo_uchar *_wcsdup (const xairo_uchar* s)
{
  size_t len = (wcslen (s) + 1) * sizeof (xairo_uchar);
  void *neww = malloc (len);

  if (neww == NULL)
    return NULL;

  return (xairo_uchar *) memcpy (neww, (void *) s, len);
}

int wcscmp (const xairo_uchar* s1,const xairo_uchar* s2)
{
  xairo_uchar c1, c2;

  do
    {
      c1 = (xairo_uchar) *s1++;
      c2 = (xairo_uchar) *s2++;
      if (c1 == L'\0')
	return c1 - c2;
    }
  while (c1 == c2);

  return c1 - c2;
}

xairo_uchar *wcschr (const xairo_uchar* wcs, const xairo_uchar wc)
{
  do
    if (*wcs == wc)
      return (xairo_uchar *) wcs;
  while (*wcs++ != L'\0');

  return NULL;
}

xairo_uchar *wcsstr (const xairo_uchar* haystack, const xairo_uchar* needle)
{
  register xairo_uchar b, c;

  if ((b = *needle) != L'\0')
    {
      haystack--;				/* possible ANSI violation */
      do
	if ((c = *++haystack) == L'\0')
	  goto ret0;
      while (c != b);

      if (!(c = *++needle))
	goto foundneedle;
      ++needle;
      goto jin;

      for (;;)
	{
	  register xairo_uchar a;
	  register const xairo_uchar *rhaystack, *rneedle;

	  do
	    {
	      if (!(a = *++haystack))
		goto ret0;
	      if (a == b)
		break;
	      if ((a = *++haystack) == L'\0')
		goto ret0;
shloop:	      ;
	    }
	  while (a != b);

jin:	  if (!(a = *++haystack))
	    goto ret0;

	  if (a != c)
	    goto shloop;

	  if (*(rhaystack = haystack-- + 1) == (a = *(rneedle = needle)))
	    do
	      {
		if (a == L'\0')
		  goto foundneedle;
		if (*++rhaystack != (a = *++needle))
		  break;
		if (a == L'\0')
		  goto foundneedle;
	      }
	    while (*++rhaystack == (a = *++needle));

	  needle = rneedle;		  /* took the register-poor approach */

	  if (a == L'\0')
	    break;
	}
    }
foundneedle:
  return (xairo_uchar*) haystack;
ret0:
  return NULL;
}

size_t mbstowcs(xairo_uchar* dest,const char* src,size_t n)
{
	size_t m=0;
	char c=1;
	while (n > 0)
    {
		c = *src++;
		*dest++ = c; m++;
		if (c == L'\0')
			return m;
		n--;
    }

	if (c != L'\0')
		*dest++ = L'\0';

	return m;
}

size_t wcstombs(char* dest,const xairo_uchar* src,size_t n)
{
	size_t m=0;
	xairo_uchar c;
	while (n > 0)
    {
		c = *src++;
		if (c >= 256) return -1;
		*dest++ = (char) c; m++;
		if (c == L'\0')
			return m;
		n--;
    }

	if (c != L'\0')
		*dest++ = L'\0';

	return m;
}

int _wtoi(const xairo_uchar* s)
{
	int x=0;
	int ineg=1;
	if (*s=='-') {
		s++;
		ineg=-1;
	}
	while (*s) {
		x=10*x+(*s)-'0';
		s++;
	}
	return x*ineg;
}


double _wtof(const xairo_uchar* s)
{
	char* c=new char [wcslen(s)+1];
	wcstombs(c,s,wcslen(s)+1);
	double f=atof(c);
	delete [] c;
	return f;
}

unsigned long wcstoul(const xairo_uchar* s,xairo_uchar** p,int r)
{
	char* c=new char [wcslen(s)+1];
	wcstombs(c,s,wcslen(s)+1);
	unsigned long l=strtoul(c,NULL,r);
	delete [] c;
	return l;
}

static xairo_uchar s_pc_d[]={'%','d',0};

xairo_uchar* _itow(int n,xairo_uchar* d,int r)
{
	swprintf(d,s_pc_d,n);
	return d;
}

// no tinylib support for wide paths

FILE* _wfopen(const xairo_uchar* path,const xairo_uchar* mode)
{
	char p[_MAX_PATH]; char m[_MAX_PATH];
	wcstombs(p,path,_MAX_PATH);
	wcstombs(m,mode,_MAX_PATH);
	return fopen(p,m);
}

int _wremove(const xairo_uchar* path)
{
	char p[_MAX_PATH];
	wcstombs(p,path,_MAX_PATH);
	return remove(p);
}

void _wrename(const xairo_uchar* path,const xairo_uchar* path1)
{
	char p[_MAX_PATH];
	char p1[_MAX_PATH];
	wcstombs(p,path,_MAX_PATH);
	wcstombs(p1,path1,_MAX_PATH);
	rename(p,p1);
}






/*
size_t wcsftime(xairo_uchar*,size_t,const xairo_uchar*,const tm*);
*/


int myswprintf(xairo_uchar* s,const xairo_uchar* fmt,va_list v)
{
	char fb[100]; char nb[100]; xairo_uchar* s0=s;
	while (*fmt) {
		if (*fmt != '%') 
		{
			*s=*fmt; fmt++; s++;
			continue;
		}
		if (*(fmt+1)=='%') 
		{
			*s='%';
			fmt+=2; s++;
			continue;
		}
		int i=0;
		fb[0]='%';
		while (*(fmt+i+1) && isdigit(*(fmt+i+1)))
		{fb[1+i]=(char) *(fmt+i+1); i++;}
		xairo_uchar fc=*(fmt+i+1);
		fb[1+i++]=(char) fc; fb[1+i]=0;
		fmt+=i+1;
		switch (fc) {
			case 's':
				{
				xairo_uchar* sa=va_arg(v,xairo_uchar*);
				wcscpy(s,sa);
				s+=wcslen(sa);
				break;
				}
			case 'S':
				{
				char* sa=va_arg(v,char*);
				mbstowcs(s,sa,strlen(sa));
				s+=strlen(sa);
				break;
				}
			default:
				{
				int x=va_arg(v,int);
				sprintf(nb,fb,x);
				mbstowcs(s,nb,strlen(nb));
				s+=strlen(nb);
				}
		}
	}
	s[0]=0;
	return s-s0;
}

bool isfc(char c)
{
	return isdigit(c)||c=='.';
}

int myswcprintf(const xairo_uchar* fmt,va_list v)
{
	char fb[100]; char nb[100]; int t=0;
	while (*fmt) {
		if (*fmt != '%') 
		{
			fmt++; t++;
			continue;
		}
		if (*(fmt+1)=='%') 
		{
			fmt+=2; t++;
			continue;
		}
		int i=0;
		fb[0]='%';
		while (*(fmt+i+1) && isfc(*(fmt+i+1)))
		{fb[1+i]=(char) *(fmt+i+1); i++;}
		xairo_uchar fc=*(fmt+i+1);
		fb[1+i++]=(char) fc; fb[1+i]=0;
		fmt+=i+1;
		switch (fc) {
			case 's':
				{
				xairo_uchar* sa=va_arg(v,xairo_uchar*);
				t+=wcslen(sa);
				break;
				}
			case 'S':
				{
				char* sa=va_arg(v,char*);
				t+=strlen(sa);
				break;
				}
			default:
				{
				int x=va_arg(v,int);
				sprintf(nb,fb,x);
				t+=strlen(nb);
				}
		}
	}
	return t;
}

int _scwprintf(const xairo_uchar* fmt,...)
{
	va_list arg;
	va_start(arg,fmt);
	int n=myswcprintf(fmt,arg);
	va_end(arg);
	return n;
}

int _vscwprintf(const xairo_uchar* fmt,va_list v)
{
	return myswcprintf(fmt,v);
}

int _vsnwprintf(xairo_uchar* dst,int n,const xairo_uchar* fmt,va_list v)
{
	va_list vv;
	va_copy(vv,v);
	int k=myswcprintf(fmt,vv);
	xairo_uchar* s=new xairo_uchar [k+1];
	myswprintf(s,fmt,v);
	if (n==-1) n=wcslen(s)+1;
	wcsncpy(dst,s,n);
	delete [] s;
	return wcslen(dst);
}

int vswprintf(xairo_uchar* dst,const xairo_uchar* f,va_list v)
{
	return _vsnwprintf(dst,-1,f,v);
}

int vfwprintf(FILE* f,const xairo_uchar* fmt,va_list v)
{
	va_list vv;
	va_copy(vv,v);
	int k=myswcprintf(fmt,vv);
	xairo_uchar* s=new xairo_uchar [k+1];
	myswprintf(s,fmt,v);
	char* ss=new char [k+1];
	wcstombs(ss,s,k+1);
	fwrite(ss,k,1,f);
	delete [] s;
	delete [] ss;
	return k;
}

int swprintf (xairo_uchar *s, const xairo_uchar *format, ...)
{
  va_list arg;
  int done;

  va_start (arg, format);
  done = vswprintf (s, format, arg);
  va_end (arg);

  return done;
}

int _snwprintf (xairo_uchar *s, size_t n, const xairo_uchar *format, ...)
{
  va_list arg;
  int done;

  va_start (arg, format);
  done = _vsnwprintf (s, n, format, arg);
  va_end (arg);

  return done;
}

int wprintf (const xairo_uchar *format, ...)
{
  va_list arg;
  int done;

  va_start (arg, format);
  done = vfwprintf (stdout, format, arg);
  va_end (arg);

  return done;
}

int fwprintf (FILE* f,const xairo_uchar *format, ...)
{
  va_list arg;
  int done;

  va_start (arg, format);
  done = vfwprintf (f, format, arg);
  va_end (arg);

  return done;
}

const xairo_uchar* _wctime(const time_t* t)
{
	return L(ctime(t));
}


void _fullpath (
        char *UserBuf,
        const char *path,
        size_t maxlen
        )
{
	if (path[0]=='/') {
		strncpy(UserBuf,path,maxlen);
		return;
	}
	getcwd(UserBuf,maxlen);
	int k=strlen(UserBuf);
	if (k+1 >= (int) maxlen) return;
	UserBuf[k]='/';
	strncpy(UserBuf+k+1,path,maxlen-k-1);
}

#endif




