// solve.cpp : Defines the entry point for the console application.
//

#define _CRT_SECURE_NO_DEPRECATE
#define _CRT_NON_CONFORMING_SWPRINTFS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>



#ifndef _WIN32
#include <readline/readline.h>
#include <readline/history.h>
#else
#include <windows.h>
#define PATH_MAX _MAX_PATH
#endif

#ifndef _WIN32
#ifndef DATAPATH
#error The symbol DATAPATH must be defined
#endif
#endif


#include <unicode/unistr.h>

#include "../XairoHeader.h"
using namespace XAIRO_NAMESPACE;

#define MAXCONTEXT 150

void Usage()
{
	printf("Usage:\n");
	printf("solve -p path\n");
	printf("     where path is the path to the corpus parameter file.\n");
	printf("solve -u host [-l port]\n");
	printf("     where host is the server host address and port the port (default 7000).\n\n");
}

void Message(UnicodeString a)
{
	char* s=new char[a.length()+1];
	a.extract(0,a.length(),s);
	printf("%s",s);
	delete [] s;
}

class MyBootLog : public CXairoIBootMonitor
{
	void Log(const xairo_uchar* u) {Message(u);}
};

int main(int argc, char* argv[])
{
	bool bErr=false;
	unsigned lport=7000;
	char paramfile[PATH_MAX];
	paramfile[0]=0;
	char url[PATH_MAX];
	url[0]=0;
	for (int i=1;i<argc;i++) {
		if (strcmp(argv[i],"-p")==0 && i+1 < argc) {
			strcpy(paramfile,argv[i+1]);
			i++;
		}
		else if (strcmp(argv[i],"-u")==0 && i+1 < argc) {
			strcpy(url,argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-l")==0 && i+1 < argc) {
			lport=atoi(argv[i+1]);
			i++;
		}
		else
			bErr=true;
	}
	if (bErr || paramfile[0]==0 && url[0]==0 || paramfile[0]!=0 && url[0]!=0 ) 
	{	
		Usage();
		return 1;
	}
	CXairoInterface* s=NULL;
	MyBootLog b;
	if (paramfile[0]) {
		CXairoIParameters xp;
		xp.parameter_file=paramfile;
		xp.bm=&b;
		xp.inth=NULL;
#ifdef _WIN32
		char tp[_MAX_PATH];
		GetTempPathA(_MAX_PATH,tp);
		xp.temppath=tp;
		char dp[_MAX_PATH];
		GetModuleFileNameA(NULL,dp,_MAX_PATH);
		char* x=strrchr(dp,'\\');
		*x=0;
		xp.datapath=dp;
#else
		xp.temppath="/tmp";
		xp.datapath=DATAPATH;
#endif
		s=CXairoImplementation::getServer(&xp);
	} else {
		s=CXairoProxyImplementation::getRemoteServer(url,lport,&b);
	}
	if (s==NULL) {
		printf("Cannot create server\n");
		return 2;
	}
	while (1) {
#ifndef _WIN32
		char* c=readline("");
		if (c==NULL) break;
		if (*c) add_history(c);
#else
		char c[1000];
		if (gets(c)==NULL) break;
#endif
		UnicodeString w(c);
#ifndef _WIN32
		free(c);
#endif
		CXairoISolution* sol=NULL;
		try {
			sol=s->solve(w.getTerminatedBuffer());
		} 
		catch (CXairoIError* e)
		{
			Message(e->getMessage());
			printf("\n");
		}
		int n=sol?sol->getCount():0;
		printf("<solveResults solutions=\"%d\">\n",n);
		for (int i=0;i<n;i++) {
			CXairoIHit* h=sol->getHit(i);
			UnicodeString x(h->getText(0,MAXCONTEXT));
			int n=x.extract(0,x.length(),NULL,"UTF-8");
			char* s=new char [n+1];
			x.extract(0,x.length(),s,"UTF-8");
			printf("%s\n",s);
			delete [] s;
		}
        printf("</solveResults>\n");

		if (sol) delete sol;
	}
	return 0;
}

