#include "stdafx.h"

#include "../XairoHeader.h"

using namespace XAIRO_NAMESPACE;

#include "XairaTestHarness.h"

class GHandle : public DefaultHandler {
		void fatalError(const SAXParseException&);
		void error(const SAXParseException&);
};

void GHandle::fatalError(const SAXParseException& e)
{
	throw e;
}

void GHandle::error(const SAXParseException& e)
{
	throw e;
}



CXairaTestHarness::CXairaTestHarness(void)
{
}

CXairaTestHarness::~CXairaTestHarness(void)
{
}

bool CXairaTestHarness::Init(const char* script)
{
	GHandle eh;
	m_parser.setErrorHandler(&eh);
	try {
		m_parser.parse(script);
	}
	catch (SAXParseException&  e ) {
		Log("Parser error in script file: %s\n",e.getMessage());		
		return false;
	}
	m_doc=m_parser.getDocument();
	return true;
}


int CXairaTestHarness::status()
{
	return m_status;
}

void CXairaTestHarness::exec(const char* script,CXairoInterface* ss,bool bCalib)
{
	m_status=0; m_nerr=0;
	m_server=ss;
// start by parsing the script
	if (!Init(script)) {
		SetStatus(1004);
		return;
	}
	if (bCalib) {
		if (!Calibrate(script)) {
			SetStatus(1006);
		}
		return;
	}
	Test();
	return;
}


void CXairaTestHarness::Log(const char* msg,...)
{
	va_list s;
	va_start(s,msg);
	vprintf(msg,s);
	va_end(s);
}

UChar s_[]={0};
UChar s_tests[]={'t','e','s','t','s',0};
UChar s_test[]={'t','e','s','t',0};
UChar s_resultset[]={'r','e','s','u','l','t','s','e','t',0};
UChar s_result[]={'r','e','s','u','l','t',0};
UChar s_max[]={'m','a','x',0};
UChar s_id[]={'i','d',0};
UChar s_query[]={'q','u','e','r','y',0};


bool CXairaTestHarness::Calibrate(const char* cal)
{
	DOMImplementation* di=DOMImplementation::getImplementation();
	m_cdoc=di->createDocument(s_,s_tests,NULL);
	DOMNodeList* l=m_doc->getElementsByTagName(s_test);
	XMLSize_t k=l->getLength();
	for (XMLSize_t i=0;i<k;i++) {
		DOMElement* t=(DOMElement*) l->item(i);
		DOMElement* r=EvalTest(t);
		CalRecord(t,r);		
	}
	DOMLSSerializer* w=di->createLSSerializer();
	DOMLSOutput* lso=di->createLSOutput();
	LocalFileFormatTarget t(cal);
	lso->setByteStream(&t);
	w->write(m_cdoc,lso);

	return true;
}

void CXairaTestHarness::Test()
{
	DOMNodeList* l=m_doc->getElementsByTagName(s_test);
	XMLSize_t k=l->getLength();
	for (XMLSize_t i=0;i<k;i++) {
		DOMElement* t=(DOMElement*) l->item(i);
		DOMElement* r=EvalTest(t);
		EvalRecord(t,r);		
	}
	if (m_nerr) m_status=1;
}


void CXairaTestHarness::CalRecord(DOMElement* test,DOMElement* result)
{
	DOMNode* c=NULL; DOMNode* cc=NULL;
	for (c=test->getFirstChild();c;c=cc) {
		cc=c->getNextSibling();
		if (c->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		if (UnicodeString(((DOMElement*) c)->getTagName()).compare(s_resultset)==0) {
			test->removeChild(c);
		}
	}
	test->appendChild(result);
	DOMElement* e=(DOMElement*) m_cdoc->importNode(test,true);
	DOMElement* r=m_cdoc->getDocumentElement();
	r->appendChild(e);
}

#define DEFMAX 5

void CXairaTestHarness::EvalRecord(DOMElement* test,DOMElement* result)
{
	DOMNodeList* l=test->getElementsByTagName(s_resultset);
	UnicodeString nn(test->getAttribute(s_max));
	UnicodeString id(test->getAttribute(s_id));
	char* ids=new char[id.length()+1];
	id.extract(0,id.length(),ids);
	UErrorCode e=U_ZERO_ERROR;
	DecimalFormat df(e);
	Formattable f;
	df.parse(nn,f,e);
	int nn1=f.getLong(&e);
	if (nn1==0) nn1=DEFMAX;
	if (l->getLength()==0) return;
	Comp((DOMElement*) l->item(0),result,nn1,ids);
	delete [] ids;
}


DOMElement* CXairaTestHarness::EvalTest(DOMElement* test)
{
	DOMNodeList* l=test->getElementsByTagName(s_query);
	if (l->getLength()==0) return NULL;
	DOMElement* query=(DOMElement*) l->item(0);
	UnicodeString q=NormSpace(query->getTextContent());
	CXairoISolution* sol=m_server->solve(q.getTerminatedBuffer());
	DOMElement* e=m_doc->createElement(s_resultset);
	if (sol==NULL) return e;
	int n=sol->getCount();
	const xairo_uchar* nn=test->getAttribute(s_max);
	UErrorCode err=U_ZERO_ERROR;
	DecimalFormat df(err);
	Formattable f;
	df.parse(nn,f,err);
	int nn1=f.getLong(&err);
	if (nn1==0) nn1=DEFMAX;
	if (n>nn1) n=nn1;
	for (int i=0;i<n;i++) {
		CXairoIHit* h=sol->getHit(i);
		DOMElement* r=m_doc->createElement(s_result);
		DOMText* t=m_doc->createCDATASection(NormSpace(h->getText(0,-1)).getTerminatedBuffer());
		r->appendChild(t);
		e->appendChild(r);
	}
	return e;
}

void CXairaTestHarness::Comp(DOMElement* e,DOMElement* ee,XMLSize_t max,const char* id)
{
	DOMNodeList* l=e->getElementsByTagName(s_result);
	DOMNodeList* ll=ee->getElementsByTagName(s_result);
	XMLSize_t n=l->getLength();
	XMLSize_t nn=ll->getLength();
	if (nn > max) nn=max;
	if (n != nn) {
		Log("Test %s:Solution counts differ\n",id);
		m_nerr++;
		return;
	}
	for (XMLSize_t i=0;i<n;i++) {
		DOMElement* t=(DOMElement*) l->item(i);
		DOMElement* tt=(DOMElement*) ll->item(i);
		UnicodeString a=NormSpace(t->getTextContent());
		UnicodeString aa=NormSpace(tt->getTextContent());
		if (a.compare(aa) != 0) {
			int na=a.length(); 
			int naa=aa.length(); 
			int m=na;
			if (m > naa) m=naa; int ni=0;
			for (ni=0;ni<m;ni++) {
				if (a.charAt(ni) != aa.charAt(ni)) break;
			}
//			UnicodeString b=a;
//			b.remove(0,na-10);
//			UnicodeString bb=aa;
//			bb.remove(0,naa-10);

			Log("Test %s/%d: Solutions differ at %d (%hd %hd)\n",id,i+1,ni,a.charAt(ni),aa.charAt(ni));
			m_nerr++;
		}
	}
}

bool iss(xairo_uchar c)
{
	return c==' '||c=='\t'||c=='\r'||c=='\n';
}

UnicodeString CXairaTestHarness::NormSpace(UnicodeString s)
{
	UnicodeString ss;
	bool bSpace=true;
	for (int i=0;i<s.length();i++) {
		UChar u=s.charAt(i);
		if (iss(u)) {
			if (!bSpace) ss+=' ';
			bSpace=true;
		}
		else {
			bSpace=false;
			ss+=u;
		}
	}
	if (bSpace && ss.length() > 0) ss.truncate(ss.length()-1);
	return ss;
}
